@@ -218,23 +218,19 @@ class _dirstatemapcommon:
         return self._docket
 
     def _read_v2_data(self):
-        data = None
         attempts = 0
         while attempts < V2_MAX_READ_ATTEMPTS:
             attempts += 1
             try:
                 # TODO: use mmap when possible
-                data = self._opener.read(self.docket.data_filename())
+                return self._opener.read(self.docket.data_filename())
             except FileNotFoundError:
                 # read race detected between docket and data file
                 # reload the docket and retry
                 self._docket = None
-        if data is None:
-            assert attempts >= V2_MAX_READ_ATTEMPTS
-            msg = b"dirstate read race happened %d times in a row"
-            msg %= attempts
-            raise error.Abort(msg)
-        return self._opener.read(self.docket.data_filename())
+        msg = b"dirstate read race happened %d times in a row"
+        msg %= attempts
+        raise error.Abort(msg)
 
     def write_v2_no_append(self, tr, st, meta, packed):
         try:
