@@ -5,15 +5,29 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
+from __future__ import annotations
+
 import collections
+import enum
 import functools
 import os
 import re
 import stat
-from typing import Generator, List
+import typing
+
+from typing import (
+    Generator,
+    Iterator,
+)
 
 from .i18n import _
 from .thirdparty import attr
+
+# Force pytype to use the non-vendored package
+if typing.TYPE_CHECKING:
+    # noinspection PyPackageRequirements
+    import attr
+
 from .node import hex
 from .revlogutils.constants import (
     INDEX_HEADER,
@@ -37,10 +51,10 @@ from .utils import hashutil
 parsers = policy.importmod('parsers')
 # how much bytes should be read from fncache in one read
 # It is done to prevent loading large fncache files into memory
-fncache_chunksize = 10 ** 6
+fncache_chunksize = 10**6
 
 
-def _match_tracked_entry(entry, matcher):
+def _match_tracked_entry(entry: BaseStoreEntry, matcher):
     """parses a fncache entry and returns whether the entry is tracking a path
     matched by matcher or not.
 
@@ -48,13 +62,56 @@ def _match_tracked_entry(entry, matcher):
 
     if matcher is None:
         return True
+
+    # TODO: make this safe for other entry types.  Currently, the various
+    #  store.data_entry generators only yield  RevlogStoreEntry, so the
+    #  attributes do exist on `entry`.
+    # pytype: disable=attribute-error
     if entry.is_filelog:
         return matcher(entry.target_id)
     elif entry.is_manifestlog:
         return matcher.visitdir(entry.target_id.rstrip(b'/'))
+    # pytype: enable=attribute-error
     raise error.ProgrammingError(b"cannot process entry %r" % entry)
 
 
+class Encoding(enum.IntEnum):
+    HYBRID = enum.auto()
+    DOTENCODE = enum.auto()
+    PLAIN = enum.auto()
+
+
+def _plain_encode(path: bytes) -> bytes:
+    """A very basic encoding that encode (almost) nothing
+
+    This only transforms filelog (or anything in 'data/store', really)
+    To avoid revlog-directory conflict, every directory will be suffixed with a
+    `_`. The filelog will use their usual `.i` and `.d` suffixes.
+
+    The change is inflicted to every directory names as this will be less
+    surprising and more future proof.
+
+    This is not an encoding that is intended for general usage, it make sense
+    for some large server installation for very controll repository.
+
+    >>> _plain_encode(b"foo")
+    b'foo'
+    >>> _plain_encode(b"foo.txt")
+    b'foo.txt'
+    >>> _plain_encode(b"bar/foo.txt")
+    b'bar/foo.txt'
+    >>> _plain_encode(b"data/foo.txt")
+    b'data/foo.txt'
+    >>> _plain_encode(b"data/bar/foo.txt")
+    b'data/bar_/foo.txt'
+    >>> _plain_encode(b"data/ b a r /fuz.d_/baz.i/foo.txt")
+    b'data/ b a r _/fuz.d__/baz.i_/foo.txt'
+    """
+    if path.startswith(b'data/'):
+        path = b'data/' + path[5:].replace(b'/', b'_/')
+    return path
+
+
 # This avoids a collision between a file named foo and a dir named
 # foo.i or foo.d
 def _encodedir(path):
@@ -96,7 +153,7 @@ def decodedir(path):
     )
 
 
-def _reserved():
+def _reserved() -> Iterator[int]:
     """characters that are problematic for filesystems
 
     * ascii escapes (0..31)
@@ -105,13 +162,10 @@ def _reserved():
 
     these characters will be escaped by encodefunctions
     """
-    winreserved = [ord(x) for x in u'\\:*?"<>|']
-    for x in range(32):
-        yield x
-    for x in range(126, 256):
-        yield x
-    for x in winreserved:
-        yield x
+    winreserved = [ord(x) for x in '\\:*?"<>|']
+    yield from range(32)
+    yield from range(126, 256)
+    yield from winreserved
 
 
 def _buildencodefun():
@@ -440,6 +494,9 @@ class StoreFile:
     unencoded_path = attr.ib()
     _file_size = attr.ib(default=None)
     is_volatile = attr.ib(default=False)
+    # Missing file can be safely ignored, used by "copy/hardlink" local clone
+    # for cache file not covered by lock.
+    optional = False
 
     def file_size(self, vfs):
         if self._file_size is None:
@@ -457,7 +514,7 @@ class StoreFile:
     def has_size(self):
         return self._file_size is not None
 
-    def get_stream(self, vfs, copies):
+    def get_stream(self, vfs, volatiles):
         """return data "stream" information for this file
 
         (unencoded_file_path, content_iterator, content_size)
@@ -465,8 +522,8 @@ class StoreFile:
         size = self.file_size(None)
 
         def get_stream():
-            actual_path = copies[vfs.join(self.unencoded_path)]
-            with open(actual_path, 'rb') as fp:
+            path = vfs.join(self.unencoded_path)
+            with volatiles.open(path) as fp:
                 yield None  # ready to stream
                 if size <= 65536:
                     yield fp.read(size)
@@ -486,14 +543,14 @@ class BaseStoreEntry:
 
     maybe_volatile = True
 
-    def files(self) -> List[StoreFile]:
+    def files(self) -> list[StoreFile]:
         raise NotImplementedError
 
     def get_streams(
         self,
         repo=None,
         vfs=None,
-        copies=None,
+        volatiles=None,
         max_changeset=None,
         preserve_file_count=False,
     ):
@@ -502,7 +559,17 @@ class BaseStoreEntry:
         return [(unencoded_file_path, content_iterator, content_size), …]
         """
         assert vfs is not None
-        return [f.get_stream(vfs, copies) for f in self.files()]
+        return [f.get_stream(vfs, volatiles) for f in self.files()]
+
+    def preserve_volatiles(self, vfs, volatiles):
+        """Use a VolatileManager to preserve the state of any volatile file
+
+        This is useful for code that need a consistent view of the content like stream clone.
+        """
+        if self.maybe_volatile:
+            for f in self.files():
+                if f.is_volatile:
+                    volatiles(vfs.join(f.unencoded_path))
 
 
 @attr.s(slots=True, init=False)
@@ -530,7 +597,7 @@ class SimpleStoreEntry(BaseStoreEntry):
         self._files = None
         self.maybe_volatile = is_volatile
 
-    def files(self) -> List[StoreFile]:
+    def files(self) -> list[StoreFile]:
         if self._files is None:
             self._files = [
                 StoreFile(
@@ -592,7 +659,7 @@ class RevlogStoreEntry(BaseStoreEntry):
         """unencoded path of the main revlog file"""
         return self._path_prefix + b'.i'
 
-    def files(self) -> List[StoreFile]:
+    def files(self) -> list[StoreFile]:
         if self._files is None:
             self._files = []
             for ext in sorted(self._details, key=_ext_key):
@@ -612,7 +679,7 @@ class RevlogStoreEntry(BaseStoreEntry):
         self,
         repo=None,
         vfs=None,
-        copies=None,
+        volatiles=None,
         max_changeset=None,
         preserve_file_count=False,
     ):
@@ -628,13 +695,13 @@ class RevlogStoreEntry(BaseStoreEntry):
             return super().get_streams(
                 repo=repo,
                 vfs=vfs,
-                copies=copies,
+                volatiles=volatiles,
                 max_changeset=max_changeset,
                 preserve_file_count=preserve_file_count,
             )
         elif not preserve_file_count:
             stream = [
-                f.get_stream(vfs, copies)
+                f.get_stream(vfs, volatiles)
                 for f in self.files()
                 if not f.unencoded_path.endswith((b'.i', b'.d'))
             ]
@@ -648,7 +715,7 @@ class RevlogStoreEntry(BaseStoreEntry):
             name_to_size[f.unencoded_path] = f.file_size(None)
 
         stream = [
-            f.get_stream(vfs, copies)
+            f.get_stream(vfs, volatiles)
             for f in self.files()
             if not f.unencoded_path.endswith(b'.i')
         ]
@@ -701,7 +768,7 @@ class RevlogStoreEntry(BaseStoreEntry):
         )
         return stream
 
-    def get_revlog_instance(self, repo):
+    def get_revlog_instance(self, repo, writable=False):
         """Obtain a revlog instance from this store entry
 
         An instance of the appropriate class is returned.
@@ -714,7 +781,7 @@ class RevlogStoreEntry(BaseStoreEntry):
                 repo.nodeconstants, repo.svfs, tree=mandir
             )
         else:
-            return filelog.filelog(repo.svfs, self.target_id)
+            return filelog.filelog(repo.svfs, self.target_id, writable=writable)
 
 
 def _gather_revlog(files_data):
@@ -761,7 +828,7 @@ def _ext_key(ext):
 class basicstore:
     '''base class for local repository stores'''
 
-    def __init__(self, path, vfstype):
+    def __init__(self, path: bytes, vfstype) -> None:
         vfs = vfstype(path)
         self.path = vfs.base
         self.createmode = _calcmode(vfs)
@@ -770,7 +837,7 @@ class basicstore:
         self.vfs = vfsmod.filtervfs(vfs, encodedir)
         self.opener = self.vfs
 
-    def join(self, f):
+    def join(self, f: bytes) -> bytes:
         return self.path + b'/' + encodedir(f)
 
     def _walk(self, relpath, recurse, undecodable=None):
@@ -803,7 +870,7 @@ class basicstore:
             concurrencychecker=concurrencychecker,
         )
 
-    def manifestlog(self, repo, storenarrowmatch):
+    def manifestlog(self, repo, storenarrowmatch) -> manifest.manifestlog:
         rootstore = manifest.manifestrevlog(repo.nodeconstants, self.vfs)
         return manifest.manifestlog(self.vfs, repo, rootstore, storenarrowmatch)
 
@@ -895,10 +962,8 @@ class basicstore:
         are passed with matches the matcher
         """
         # yield data files first
-        for x in self.data_entries(matcher):
-            yield x
-        for x in self.top_entries(phase=phase, obsolescence=obsolescence):
-            yield x
+        yield from self.data_entries(matcher)
+        yield from self.top_entries(phase=phase, obsolescence=obsolescence)
 
     def copylist(self):
         return _data
@@ -925,7 +990,7 @@ class basicstore:
 
 
 class encodedstore(basicstore):
-    def __init__(self, path, vfstype):
+    def __init__(self, path: bytes, vfstype) -> None:
         vfs = vfstype(path + b'/store')
         self.path = vfs.base
         self.createmode = _calcmode(vfs)
@@ -953,14 +1018,12 @@ class encodedstore(basicstore):
     def data_entries(
         self, matcher=None, undecodable=None
     ) -> Generator[BaseStoreEntry, None, None]:
-        entries = super(encodedstore, self).data_entries(
-            undecodable=undecodable
-        )
+        entries = super().data_entries(undecodable=undecodable)
         for entry in entries:
             if _match_tracked_entry(entry, matcher):
                 yield entry
 
-    def join(self, f):
+    def join(self, f: bytes) -> bytes:
         return self.path + b'/' + encodefilename(f)
 
     def copylist(self):
@@ -978,12 +1041,16 @@ class fncache:
         # set of new additions to fncache
         self.addls = set()
 
+    @property
+    def is_loaded(self):
+        return self.entries is not None
+
     def ensureloaded(self, warn=None):
         """read the fncache file if not already read.
 
         If the file on disk is corrupted, raise. If warn is provided,
         warn and keep going instead."""
-        if self.entries is None:
+        if not self.is_loaded:
             self._load(warn)
 
     def _load(self, warn=None):
@@ -991,7 +1058,7 @@ class fncache:
         self._dirty = False
         try:
             fp = self.vfs(b'fncache', mode=b'rb')
-        except IOError:
+        except OSError:
             # skip nonexistent file
             self.entries = set()
             return
@@ -1038,7 +1105,7 @@ class fncache:
 
     def write(self, tr):
         if self._dirty:
-            assert self.entries is not None
+            assert self.is_loaded
             self.entries = self.entries | self.addls
             self.addls = set()
             tr.addbackup(b'fncache')
@@ -1063,13 +1130,13 @@ class fncache:
     def add(self, fn):
         if fn in self._ignores:
             return
-        if self.entries is None:
+        if not self.is_loaded:
             self._load()
         if fn not in self.entries:
             self.addls.add(fn)
 
     def remove(self, fn):
-        if self.entries is None:
+        if not self.is_loaded:
             self._load()
         if fn in self.addls:
             self.addls.remove(fn)
@@ -1083,12 +1150,12 @@ class fncache:
     def __contains__(self, fn):
         if fn in self.addls:
             return True
-        if self.entries is None:
+        if not self.is_loaded:
             self._load()
         return fn in self.entries
 
     def __iter__(self):
-        if self.entries is None:
+        if not self.is_loaded:
             self._load()
         return iter(self.entries | self.addls)
 
@@ -1096,8 +1163,10 @@ class fncache:
 class _fncachevfs(vfsmod.proxyvfs):
     def __init__(self, vfs, fnc, encode):
         vfsmod.proxyvfs.__init__(self, vfs)
-        self.fncache = fnc
+        self.fncache: fncache = fnc
         self.encode = encode
+        self.uses_dotencode = encode is _pathencode
+        self.uses_plain_encode = encode is _plain_encode
 
     def __call__(self, path, mode=b'r', *args, **kw):
         encoded = self.encode(path)
@@ -1108,7 +1177,7 @@ class _fncachevfs(vfsmod.proxyvfs):
         ):
             # do not trigger a fncache load when adding a file that already is
             # known to exist.
-            notload = self.fncache.entries is None and (
+            notload = not self.fncache.is_loaded and (
                 # if the file has size zero, it should be considered as missing.
                 # Such zero-size files are the result of truncation when a
                 # transaction is aborted.
@@ -1119,11 +1188,13 @@ class _fncachevfs(vfsmod.proxyvfs):
                 self.fncache.add(path)
         return self.vfs(encoded, mode, *args, **kw)
 
-    def join(self, path):
+    def join(self, path: bytes | None, *insidef: bytes) -> bytes:
+        insidef = (self.encode(f) for f in insidef)
+
         if path:
-            return self.vfs.join(self.encode(path))
+            return self.vfs.join(self.encode(path), *insidef)
         else:
-            return self.vfs.join(path)
+            return self.vfs.join(path, *insidef)
 
     def register_file(self, path):
         """generic hook point to lets fncache steer its stew"""
@@ -1132,11 +1203,15 @@ class _fncachevfs(vfsmod.proxyvfs):
 
 
 class fncachestore(basicstore):
-    def __init__(self, path, vfstype, dotencode):
-        if dotencode:
+    def __init__(self, path: bytes, vfstype, encoding: Encoding) -> None:
+        if encoding == Encoding.DOTENCODE:
             encode = _pathencode
-        else:
+        elif encoding == Encoding.HYBRID:
             encode = _plainhybridencode
+        elif encoding == Encoding.PLAIN:
+            encode = _plain_encode
+        else:
+            assert False, encoding
         self.encode = encode
         vfs = vfstype(path + b'/store')
         self.path = vfs.base
@@ -1149,7 +1224,7 @@ class fncachestore(basicstore):
         self.vfs = _fncachevfs(vfs, fnc, encode)
         self.opener = self.vfs
 
-    def join(self, f):
+    def join(self, f: bytes) -> bytes:
         return self.pathsep + self.encode(f)
 
     def getsize(self, path):
